#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <X11/xpm.h>
#include <X11/Xatom.h>
#include <Imlib2.h>

#include "../wmgeneral/wmgeneral.h"
#include "wmconfig.h"

//globals from wmgeneral or wmmsg.h
extern Display *display;
extern GC NormalGC;
extern XpmIcon wmgen;
extern int global_intLocation_x[];
extern int global_intLocation_y[];
extern char verbose;

//classwide variables
static char *names[] = DEFAULT_PROTOCOL_NAMES;
static char *protocol_actions[] = DEFAULT_PROTOCOL_ACTIONS;

//fun headers
static int ReadConfigString(FILE *, char *, char *, int);
static int ReadConfigInt(FILE *, char *, int *);
static void load_images(char *[], struct config_data *);
static int getDefaultIconIndex(char *);
static void outOfMemory(void);

/*******************************************************
 * Initialize a new config_data struct and load settings
 * from the config file filename into it.
 *******************************************************/
struct config_data * get_new_config_data(char * filename)
{
    struct config_data * data;
    FILE *fp;
    int seek = 0, tempseek;
    char buf[512];
    int intCurrProtocol;
    //char *protocol_icon_paths[MAX_PROTOCOLS_ALLOWED];
    char **protocol_icon_paths;
    int i;

    if(verbose) fprintf(stderr, "getting new config data\n");

    /*****************************************
    * first, initialize everything to default
    *****************************************/
    data = malloc(sizeof(struct config_data));
    if(!data) outOfMemory();
    data->use_global_command = DEFAULT_USE_GLOBAL_COMMAND;
    data->global_command = DEFAULT_GLOBAL_COMMAND;
    data->max_messages = DEFAULT_MAX_MESSAGES;
    data->time_format = DEFAULT_TIME_FORMAT;
    data->grouped_icons = DEFAULT_GROUPED_ICONS;
    data->click_clears_all = DEFAULT_CLICK_CLEARS_ALL;
    data->blink_enabled = DEFAULT_BLINK_ENABLED;
    data->number_of_blinks = DEFAULT_NUMBER_OF_BLINKS;
    data->monitor_desktops = DEFAULT_MONITOR_DESKTOPS;
    data->msg_desktop_num = DEFAULT_MSG_DESKTOP_NUM;
    data->protocol_count = DEFAULT_PROTOCOL_COUNT;
    data->protocol_names = names;
    data->protocol_actions = protocol_actions;

    protocol_icon_paths = calloc(MAX_PROTOCOLS_ALLOWED, sizeof(char*));
    if(!protocol_icon_paths) outOfMemory();

    /********************************
    * open config file and load data
    ********************************/
    if(verbose) fprintf(stderr, "opening config file \"%s\"\n", filename);
    fp = fopen(filename, "r");
    if(fp)
    {
        if(verbose) fprintf(stderr, "reading ints\n");
        ReadConfigInt(fp, "use_global_command", &data->use_global_command);
        ReadConfigInt(fp, "max_messages", &data->max_messages);
        ReadConfigInt(fp, "grouped_icons", &data->grouped_icons);
        ReadConfigInt(fp, "click_clears_all", &data->click_clears_all);
        ReadConfigInt(fp, "blink_enabled", &data->blink_enabled);
        ReadConfigInt(fp, "number_of_blinks", &data->number_of_blinks);
        ReadConfigInt(fp, "monitor_desktops", &data->monitor_desktops);
        ReadConfigInt(fp, "msg_desktop_num", &data->msg_desktop_num);
        data->msg_desktop_num -= 1; //convert base 1 to base 0

        if(verbose) fprintf(stderr, "done reading ints, reading strings\n");

        buf[0] = '\0';
        ReadConfigString(fp, "time_format", buf, 0);
        if(buf[0])
        {
            data->time_format = malloc(sizeof(char)*strlen(buf)+1);
            if(!data->time_format) outOfMemory();
            strncpy(data->time_format, buf, strlen(buf)+1);
        }

        buf[0] = '\0';
        ReadConfigString(fp, "global_command", buf, 0);
        if(buf[0])
        {
            data->global_command = malloc(sizeof(char)*strlen(buf)+1);
            if(!data->global_command) outOfMemory();
            strncpy(data->global_command, buf, strlen(buf)+1);
        }

        buf[0] = '\0';
        ReadConfigString(fp, "unknown_icon", buf, 0);
        if(buf[0])
        {
            protocol_icon_paths[0] = malloc(sizeof(char)*strlen(buf)+1);
            if(!protocol_icon_paths[0]) outOfMemory();
            strncpy(protocol_icon_paths[0], buf, strlen(buf)+1);
        }

        /***************************
        * read protocol definitions
        ***************************/
        data->protocol_names = malloc(sizeof(char*)*MAX_PROTOCOLS_ALLOWED);
        if(!data->protocol_names) outOfMemory();
        data->protocol_names[0] = "unknown";

        data->protocol_actions = malloc(sizeof(char*)*MAX_PROTOCOLS_ALLOWED);
        if(!data->protocol_actions) outOfMemory();
        data->protocol_actions[0] = "";

        intCurrProtocol = 1; //unknown protocols are index 0, so start at 1
        while(seek != -1 && intCurrProtocol < MAX_PROTOCOLS_ALLOWED)
        {
            if(verbose) fprintf(stderr, "looping, %d\n", intCurrProtocol);

            tempseek = ReadConfigString(fp, "protocol_name", buf, seek);
            if(tempseek)
            {
                seek += tempseek;

                //name
                if(verbose) fprintf(stderr, "buf: %s\n", buf);
                data->protocol_names[intCurrProtocol] =
                    malloc(sizeof(char)*strlen(buf)+1);
                if(!data->protocol_names[intCurrProtocol]) outOfMemory();
                strncpy(data->protocol_names[intCurrProtocol], buf,
                    strlen(buf)+1);

                //icon
                tempseek = ReadConfigString(fp, "protocol_icon", buf, seek);
                if(tempseek)
                {
                    seek += tempseek;
                    protocol_icon_paths[intCurrProtocol] =
                        malloc(sizeof(char)*strlen(buf)+1);
                    if(!protocol_icon_paths[intCurrProtocol]) outOfMemory();
                }
                if(verbose) fprintf(stderr, "buf: %s\n", buf);
                strncpy(protocol_icon_paths[intCurrProtocol], buf,
                    strlen(buf)+1);

                //action

                tempseek = ReadConfigString(fp, "protocol_action", buf, seek);
                if(tempseek)
                {
                    seek += tempseek;
                    data->protocol_actions[intCurrProtocol] =
                        malloc(sizeof(char)*strlen(buf)+1);
                    if(!data->protocol_actions[intCurrProtocol]) outOfMemory();

                    if(verbose) fprintf(stderr, "buf: %s\n", buf);
                    strncpy(data->protocol_actions[intCurrProtocol], buf,
                        strlen(buf)+1);
                }
                else
                {
                    data->protocol_actions[intCurrProtocol] = 0;
                }

                intCurrProtocol++;
            }
            else
            {
                break;
            }
        }
        if(verbose) fprintf(stderr, "done reading config file\n");
        data->protocol_count = intCurrProtocol;
        fclose(fp);
    }

    if(verbose) fprintf(stderr, "loading icons\n");
    load_images(protocol_icon_paths, data);
    if(verbose) fprintf(stderr, "done loading icons\n");

    //free icon paths.
    for(i=0; i<MAX_PROTOCOLS_ALLOWED; i++)
    {
        if(protocol_icon_paths[i])
            free(protocol_icon_paths[i]);
    }
    free(protocol_icon_paths);

    return data;
}

/************************************************
 * Returns the index of the protocol that matches
 * the passed string msg.
 ************************************************/
int match_protocol(struct config_data *data, char *msg)
{
    int i, len;

    for(i = 0; i < data->protocol_count; i++)
    {
        len = strlen(data->protocol_names[i]);
        if(len != strlen(msg)) continue;
        if(strncasecmp(msg, data->protocol_names[i], len)==0)
        {
            return i;
        }
    }
    return 0; // 0 is for unknown protocol
}

/***********************************************
* Reads a particular string from the config file
* into value.
*
* returns: the seek location immediately after
* the string that was read.
************************************************/
int ReadConfigString(FILE *fp, char *setting, char *value, int seek)
{
    char var[1024];
    char buf[1024];
    int i;
    int buflen;
    int varlen;
    int loc;

    if (!fp)
    {
        return 0;
    }

    sprintf(var, "%s=", setting);
    varlen = strlen(var);

    fseek(fp, seek, SEEK_SET);

    //for each line of the config file...
    for(loc=0; feof(fp)==0; loc+=buflen+1)
    {
        //read line into buf
        if (!fgets(buf, 512, fp))
            break;

        buflen = strlen(buf) - 1;

        //chop off newline
        buf[buflen] = '\0';

        //does it match?
        if ( strncmp(buf, var, varlen) == 0)
        {
            strncpy(value, buf + varlen, buflen - varlen + 1);
            return loc + buflen + 1;
        }
    }

    return 0;
}

/************************************************
* Reads a particular integer from the config file
*************************************************/
int ReadConfigInt(FILE *fp, char *setting, int *value)
{
    char buf[1024];

    if (ReadConfigString(fp, setting, (char *) &buf, 0))
    {
        *value = atoi(buf);
        return 1;
    }

    return 0;
}

/******************************************
* Renders the icons to Pixmaps using imlib.
*******************************************/
void load_images(char * locations[], struct config_data *data)
{
    //location of icons in XPM file
    int default_icon;

    //default screen of X server
    int screen;

    //default Visual of X server
    Visual * vis;

    //default Colormap of X server
    Colormap cm;

    //temporary imlib pointer for icon loading
    Imlib_Image icon;

    //loop index
    int i;

    // set the default X stuff
    screen = DefaultScreen(display);
    vis = DefaultVisual(display, screen);
    cm = DefaultColormap(display, screen);

    // set up imlib options
    imlib_context_set_display(display);
    imlib_context_set_visual(vis);
    imlib_context_set_colormap(cm);

    // set unknown icon
    data->protocol_pixmaps[0] = XCreatePixmap(display,
        RootWindow(display, screen), 14, 13, DefaultDepth(display, screen));
    XCopyArea (display, wmgen.pixmap, data->protocol_pixmaps[0],
        NormalGC, global_intLocation_x[0], global_intLocation_y[0],
        14, 13, 0, 0);

    // initialize Pixmaps and render icons
    for(i = 0; i < data->protocol_count; i++)
    {
        data->protocol_pixmaps[i] = XCreatePixmap(display, RootWindow(display, screen), 14, 13, DefaultDepth(display, screen));

        // if there is no filename or the filename is "default",
        // then copy the default icon from the global pixmap
        if ((!locations[i]) ||
            (locations[i] && strlen(locations[i]) == 7 &&
            strncasecmp(locations[i], "default", 7) == 0))
        {
            default_icon = getDefaultIconIndex(data->protocol_names[i]);
            XCopyArea (display, wmgen.pixmap, data->protocol_pixmaps[i],
                NormalGC, global_intLocation_x[default_icon],
                global_intLocation_y[default_icon], 14, 13, 0, 0);
        }
        else if(locations[i])
        {
            // load the specified icon
            icon = imlib_load_image(locations[i]);
            if(!icon)
            {
                fprintf(stderr, "ERROR LOADING IMAGE %s\n", locations[i]);
                continue;
            }

            // copy the background
            XCopyArea (display, wmgen.pixmap, data->protocol_pixmaps[i],
                NormalGC, 5, 5, 14, 13, 0, 0);
            XCopyArea (display, wmgen.pixmap, data->protocol_pixmaps[i],
                NormalGC, 5, 5, 14, 13, 0, 0);

            // render the icon to Pixmap
            imlib_context_set_image(icon);
            imlib_context_set_drawable(data->protocol_pixmaps[i]);
            imlib_render_image_on_drawable_at_size(0, 0, 14, 13);
            imlib_free_image();
        }
    }
}

int getDefaultIconIndex(char *name)
{
    int i, len;

    for(i = 0; i < DEFAULT_PROTOCOL_COUNT; i++)
    {
        len = strlen(name);
        if(len != strlen(names[i])) continue;
        if(strncasecmp(name, names[i], len)==0)
        {
            return i;
        }
    }
    return 0;
}

//prints out of memory error and exits
void outOfMemory()
{
    fprintf(stderr, "Error: wmmsg out of memory!\n");
    _exit(1);
}
